/*
 * IWS520-USB Control Program (Linux)
 * Copyright (C) 2017 Tokyo Devices, I.W. Technology Firm, Inc.
 * https://tokyodevices.jp/
 * License: GNU GPL v2 (see License.txt)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include "hiddata.h"

#define USB_CFG_VENDOR_ID       0xc0, 0x16
#define USB_CFG_DEVICE_ID       0xdf, 0x05
#define USB_CFG_VENDOR_NAME     'i', 'w', '-', 't', 'e', 'c', 'h', 'f', 'i', 'r', 'm' , '.', 'c', 'o', 'm'
#define USB_CFG_DEVICE_NAME     'I', 'W', 'S' , '5', '2', '0'

#define FMV 0x010c
#define CLOCK_FREQ 187500
#define OOR_DIST 9999
#define LISTEN_LEN 6500

usbDevice_t *dev;
char buffer[17];

static char *usbErrorMessage(int errCode)
{
	static char buffer[80];

	switch(errCode)
	{
		case USBOPEN_ERR_ACCESS:      return "Access to device denied";
		case USBOPEN_ERR_NOTFOUND:    return "The specified device was not found";
		case USBOPEN_ERR_IO:          return "Communication error with device";
		default:
			sprintf(buffer, "Unknown USB error %d", errCode);
			return buffer;
	}

	return NULL;    /* not reached */
}

static usbDevice_t  *openDevice(char *serialNumber)
{
  usbDevice_t     *dev = NULL;
  unsigned char   rawVid[2] = {USB_CFG_VENDOR_ID};
  unsigned char   rawPid[2] = {USB_CFG_DEVICE_ID};
  char            vendorName[] = {USB_CFG_VENDOR_NAME, 0};
  char		  productName[] = {USB_CFG_DEVICE_NAME, 0};
  int             vid = rawVid[0] + 256 * rawVid[1];
  int             pid = rawPid[0] + 256 * rawPid[1];	
  int             err;

  if((err = usbhidOpenDevice(&dev, vid, vendorName, pid, productName, serialNumber, 0)) != 0){
    if( serialNumber != NULL ) {
      fprintf(stderr, "error finding %s: %s\n", serialNumber, usbErrorMessage(err));
    }
    return NULL;
  }
  
  return dev;
}

static void usage(char *myName)
{
  printf("IWS520-USB host controller (Linux) Version 0.1.1\n");
  printf("2014-2017 (C) I.W. Technology Firm, Inc., Tokyo Devices\n");
  printf("https://tokyodevices.jp/\n");
  printf("Usage: %s <MODE> <SERIAL> <COUNT> <WAIT> <SPEED> <EMIT_LEN> <MASK_DIST>\n",myName);
  printf("  <MODE>\t \"MEASURE\" \"LIST\"\n");
  printf("  <SERIAL>\t SerialNumber or \"ANY\"\n");
  printf("  <COUNT>\t Mesurement count (0 for inifinity).\n");
  printf("  <WAIT>\t Delay time for next mesurement in msec.\n");
  printf("  <SPEED>\t Speed of sound in m/sec.\n");
  printf("  <EMIT_LEN>\t Time length for ultra-sonic emission in clock.\n");
  printf("  <MASK_DIST>\t Distance to ignore the echos in meter.\n");
}

static int measure(char **argv)
{
  int err;
  double SPEED;
  int COUNT;
  int WAIT;
  int EMIT_LEN;
  int MASK_LEN;
  double MASK_DIST;
  int counter;
  double m;
  int recv_len = sizeof(buffer);
  
  COUNT = atoi(argv[3]);
  if( COUNT < 0 ) {
    fprintf(stderr, "COUNT must greater than or equal to 0.\n");
    return -1;
  }

  WAIT = atoi(argv[4]);
  if( WAIT < 75 ) {
    fprintf(stderr, "WAIT must be greater than or equal to 75msec.\n");
    return -1;
  }

  SPEED = atof(argv[5]);

  EMIT_LEN = atoi(argv[6]);
  if( EMIT_LEN < 8 || EMIT_LEN > 188 ) {
    fprintf(stderr, "EMIT_LEN must be in 8 to 188.\n");
    return -1;
  }

  MASK_DIST = atof(argv[7]);
  if( MASK_DIST < 0.0 || MASK_DIST > 5.0 ) {
    fprintf(stderr, "MASK_DIST must be in 0.0 to 5.0.\n");
    return -1;
  }

  MASK_LEN = (int)(((double)CLOCK_FREQ * 2.0 * MASK_DIST ) / SPEED);

  if((dev = openDevice(argv[2])) == NULL) return -1;

  memset(buffer, 0, sizeof(buffer));
  buffer[0] = 0;
  buffer[1] = 0x81;
  buffer[2] = (EMIT_LEN >> 8); buffer[3] = (EMIT_LEN & 0xFF);
  buffer[4] = (MASK_LEN >> 8); buffer[5] = (MASK_LEN & 0xFF);
  buffer[6] = (LISTEN_LEN >> 8); buffer[7] = (LISTEN_LEN & 0xFF);

  if((err = usbhidSetReport(dev, buffer, sizeof(buffer))) != 0) {
    fprintf(stderr, "error writing parameters: %s\n", usbErrorMessage(err));
    return -2;
  }

  counter = 0;
  while(1) {

    if((err = usbhidGetReport(dev, 0, buffer, &recv_len)) != 0 ) {
      printf("\"ERR\",,\n");
    } else {
      int d = (((unsigned char)buffer[2] << 8) | (unsigned char)buffer[1] );
      int t = (((unsigned char)buffer[4] << 8) | (unsigned char)buffer[3] );
      int f = (((unsigned char)buffer[6] << 8) | (unsigned char)buffer[5] );
      int n = (((unsigned char)buffer[8] << 8) | (unsigned char)buffer[7] );
      
      if( f != FMV ) d = d + MASK_LEN + 38;

      if( d == OOR_DIST ) {
	printf("\"OOR\",%d,%d\n", t, n);
      } else {
	m = (64.0 / 12000000.0) * ((double)d / 2.0) * SPEED;
	printf("%.3f,%d,%d\n", (double)m, t, n);
      }
      
    }
    
    fflush(stdout);

    counter++;
    if( COUNT > 0 && counter == COUNT ) break;
    
    usleep(WAIT * 1000);
  }

  return 0;
}

int main(int argc, char **argv)
{
  
  if(argc < 2) {
    usage(argv[0]);
    exit(1);
  }
  
  if(strcmp(argv[1], "LIST") == 0) {
    
    if ((dev = openDevice(NULL)) == NULL ) exit(1);

  } else if( strcmp(argv[1], "MEASURE") == 0  ){

    if( argc < 8 ) {
      fprintf(stderr, "Invalid arguments.\n");
      usage(argv[0]);
      exit(1);
    }

    measure(argv);
     
  } else {
    usage(argv[0]);
    exit(1);
  }

  usbhidCloseDevice(dev);
	
  return 0;
}
